@@ -18,6 +18,7 @@ import inspect
 
 import argparse
 import mock
+import pkg_resources
 import unittest2
 
 import cli_tools
@@ -56,16 +57,25 @@ class TestScriptAdaptor(unittest2.TestCase):
         self.assertEqual(func._script_adaptor, result)
         self.assertEqual(func.args_hook, result.args_hook)
         self.assertEqual(func.processor, result.processor)
+        self.assertEqual(func.subcommand, result.subcommand)
         self.assertEqual(func.setup_args, result.setup_args)
         self.assertEqual(func.get_kwargs, result.get_kwargs)
         self.assertEqual(func.safe_call, result.safe_call)
         self.assertEqual(func.console, result.console)
+        self.assertEqual(func.get_subcommands, result.get_subcommands)
 
     def test_get_adaptor_set(self):
-        func = mock.Mock(__doc__='', args_hook='args_hook',
-                         processor='processor', setup_args='setup_args',
-                         get_kwargs='get_kwargs', safe_call='safe_call',
-                         console='console')
+        func = mock.Mock(
+            __doc__='',
+            args_hook='args_hook',
+            processor='processor',
+            subcommand='subcommand',
+            setup_args='setup_args',
+            get_kwargs='get_kwargs',
+            safe_call='safe_call',
+            console='console',
+            get_subcommands='get_subcommands',
+        )
         sa = cli_tools.ScriptAdaptor(func)
         func._script_adaptor = sa
 
@@ -74,10 +84,12 @@ class TestScriptAdaptor(unittest2.TestCase):
         self.assertEqual(result, sa)
         self.assertEqual(func.args_hook, 'args_hook')
         self.assertEqual(func.processor, 'processor')
+        self.assertEqual(func.subcommand, 'subcommand')
         self.assertEqual(func.setup_args, 'setup_args')
         self.assertEqual(func.get_kwargs, 'get_kwargs')
         self.assertEqual(func.safe_call, 'safe_call')
         self.assertEqual(func.console, 'console')
+        self.assertEqual(func.get_subcommands, 'get_subcommands')
 
     def test_init(self):
         func = mock.Mock(__doc__="description")
@@ -90,11 +102,15 @@ class TestScriptAdaptor(unittest2.TestCase):
         self.assertEqual(sa._processor('foo'), None)
         self.assertEqual(sa._arguments, [])
         self.assertEqual(sa._groups, {})
+        self.assertEqual(sa._subcommands, {})
+        self.assertEqual(sa.do_subs, False)
+        self.assertEqual(sa.subkwargs, {})
         self.assertEqual(sa.prog, None)
         self.assertEqual(sa.usage, None)
         self.assertEqual(sa.description, 'description')
         self.assertEqual(sa.epilog, None)
         self.assertEqual(sa.formatter_class, argparse.HelpFormatter)
+        self.assertEqual(sa._subcmd_attr, '_script_adaptor_%x' % id(sa))
 
     def test_add_argument(self):
         func = mock.Mock(__doc__='')
@@ -138,6 +154,44 @@ class TestScriptAdaptor(unittest2.TestCase):
         self.assertRaises(argparse.ArgumentError, sa._add_group,
                           'group', 'group', dict(a=1, b=2, c=3))
 
+    def test_add_subcommand(self):
+        func = mock.Mock(__doc__='')
+        sa = cli_tools.ScriptAdaptor(func)
+
+        sa._add_subcommand('cmd', 'adaptor')
+        sa._add_subcommand('dmc', 'rotpada')
+
+        self.assertEqual(sa._subcommands, dict(cmd='adaptor', dmc='rotpada'))
+        self.assertEqual(sa.do_subs, True)
+
+    @mock.patch.object(pkg_resources, 'iter_entry_points', return_value=[
+        mock.Mock(**{'load.side_effect': ImportError}),
+        mock.Mock(**{'load.side_effect': pkg_resources.UnknownExtra}),
+        mock.Mock(**{'load.side_effect': AttributeError}),
+        mock.Mock(**{
+            'load.return_value': mock.Mock(_script_adaptor='adaptor'),
+        }),
+        mock.Mock(**{
+            'load.return_value': mock.Mock(_script_adaptor='rotpada'),
+        }),
+    ])
+    @mock.patch.object(cli_tools.ScriptAdaptor, '_add_subcommand')
+    def test_add_extensions(self, mock_add_subcommand, mock_iter_entry_points):
+        mock_iter_entry_points.return_value[-2].name = 'cmd'
+        mock_iter_entry_points.return_value[-1].name = 'dmc'
+
+        func = mock.Mock(__doc__='')
+        sa = cli_tools.ScriptAdaptor(func)
+
+        sa._add_extensions('entrypoint.group')
+
+        mock_iter_entry_points.assert_called_once_with('entrypoint.group')
+        mock_add_subcommand.assert_has_calls([
+            mock.call('cmd', 'adaptor'),
+            mock.call('dmc', 'rotpada'),
+        ])
+        self.assertEqual(sa.do_subs, True)
+
     def test_args_hook(self):
         func = mock.Mock(__doc__='')
         sa = cli_tools.ScriptAdaptor(func)
@@ -156,6 +210,58 @@ class TestScriptAdaptor(unittest2.TestCase):
         self.assertEqual(result, 'func')
         self.assertEqual(sa._processor, 'func')
 
+    @mock.patch.object(cli_tools.ScriptAdaptor, '_get_adaptor',
+                       return_value='adaptor')
+    @mock.patch.object(cli_tools.ScriptAdaptor, '_add_subcommand')
+    def test_subcommand_basic(self, mock_add_subcommand, mock_get_adaptor):
+        func = mock.Mock(__doc__='')
+        sa = cli_tools.ScriptAdaptor(func)
+
+        decorator = sa.subcommand('cmd')
+
+        self.assertTrue(callable(decorator))
+        self.assertFalse(mock_add_subcommand.called)
+
+        subcmd = mock.Mock(__name__='subcmd')
+        result = decorator(subcmd)
+
+        self.assertEqual(result, subcmd)
+        mock_get_adaptor.assert_called_once_with(subcmd)
+        mock_add_subcommand.assert_called_once_with('cmd', 'adaptor')
+
+    @mock.patch.object(cli_tools.ScriptAdaptor, '_get_adaptor',
+                       return_value='adaptor')
+    @mock.patch.object(cli_tools.ScriptAdaptor, '_add_subcommand')
+    def test_subcommand_derived(self, mock_add_subcommand, mock_get_adaptor):
+        func = mock.Mock(__doc__='')
+        sa = cli_tools.ScriptAdaptor(func)
+
+        decorator = sa.subcommand()
+
+        self.assertTrue(callable(decorator))
+        self.assertFalse(mock_add_subcommand.called)
+
+        subcmd = mock.Mock(__name__='subcmd')
+        result = decorator(subcmd)
+
+        self.assertEqual(result, subcmd)
+        mock_get_adaptor.assert_called_once_with(subcmd)
+        mock_add_subcommand.assert_called_once_with('subcmd', 'adaptor')
+
+    @mock.patch.object(cli_tools.ScriptAdaptor, '_get_adaptor',
+                       return_value='adaptor')
+    @mock.patch.object(cli_tools.ScriptAdaptor, '_add_subcommand')
+    def test_subcommand_noparams(self, mock_add_subcommand, mock_get_adaptor):
+        func = mock.Mock(__doc__='')
+        sa = cli_tools.ScriptAdaptor(func)
+        subcmd = mock.Mock(__name__='subcmd')
+
+        result = sa.subcommand(subcmd)
+
+        self.assertEqual(result, subcmd)
+        mock_get_adaptor.assert_called_once_with(subcmd)
+        mock_add_subcommand.assert_called_once_with('subcmd', 'adaptor')
+
     @mock.patch.object(inspect, 'isgeneratorfunction', return_value=False)
     @mock.patch.object(inspect, 'isgenerator', return_value=False)
     def test_setup_args(self, mock_isgenerator, mock_isgeneratorfunction):
@@ -512,6 +618,115 @@ class TestScriptAdaptor(unittest2.TestCase):
         ])
         self.assertEqual(len(sa._args_hook.return_value.method_calls), 3)
 
+    @mock.patch.object(inspect, 'isgeneratorfunction', return_value=False)
+    @mock.patch.object(inspect, 'isgenerator', return_value=False)
+    def test_setup_args_subcmds(self, mock_isgenerator,
+                                mock_isgeneratorfunction):
+        cmd_parser = mock.Mock(name='cmd')
+        dmc_parser = mock.Mock(name='dmc')
+        parser = mock.Mock(**{
+            'add_subparsers.return_value': mock.Mock(**{
+                'add_parser.side_effect': [cmd_parser, dmc_parser],
+            }),
+        })
+        func = mock.Mock(__doc__='')
+        sa = cli_tools.ScriptAdaptor(func)
+        sa._groups = {
+            'group_key': {
+                'type': 'group',
+                'arguments': [
+                    ((1, 2, 3), dict(a=4, b=5, c=6)),
+                    ((2, 3, 4), dict(a=5, b=6, c=7)),
+                ],
+            },
+            'exclusive_key': {
+                'type': 'exclusive',
+                'arguments': [
+                    ((3, 4, 5), dict(a=6, b=7, c=8)),
+                    ((4, 5, 6), dict(a=7, b=8, c=9)),
+                ],
+            },
+            'other_key': {
+                'type': 'other',
+                'arguments': [
+                    ((5, 6, 7), dict(a=8, b=9, c=0)),
+                    ((6, 7, 8), dict(a=9, b=0, c=1)),
+                ],
+            },
+        }
+        sa._arguments = [
+            ('argument', (7, 8, 9), dict(a=0, b=1, c=2)),
+            ('argument', (8, 9, 0), dict(a=1, b=2, c=3)),
+            ('group', 'group_key', dict(title='title', description='desc')),
+            ('argument', (9, 0, 1), dict(a=2, b=3, c=4)),
+            ('group', 'exclusive_key', dict(required=True)),
+            ('group', 'other_key', dict(something='nothing')),
+            ('argument', (0, 1, 2), dict(a=3, b=4, c=5)),
+            ('other', 'args', 'kwargs'),
+        ]
+        cmd_adaptor = mock.Mock(
+            prog='cmd_prog',
+            usage='cmd_usage',
+            description='cmd_description',
+            epilog='cmd_epilog',
+            formatter_class='cmd_formatter_class',
+        )
+        dmc_adaptor = mock.Mock(
+            prog='dmc_prog',
+            usage='dmc_usage',
+            description='dmc_description',
+            epilog='dmc_epilog',
+            formatter_class='dmc_formatter_class',
+        )
+        sa._subcommands = mock.Mock(**{'items.return_value': [
+            ('cmd', cmd_adaptor),
+            ('dmc', dmc_adaptor),
+        ]})
+        sa.do_subs = True
+        sa.subkwargs = dict(a=1, b=2, c=3)
+
+        sa.setup_args(parser)
+
+        parser.assert_has_calls([
+            mock.call.add_argument(7, 8, 9, a=0, b=1, c=2),
+            mock.call.add_argument(8, 9, 0, a=1, b=2, c=3),
+            mock.call.add_argument_group(title='title', description='desc'),
+            mock.call.add_argument_group()
+                .add_argument(1, 2, 3, a=4, b=5, c=6),
+            mock.call.add_argument_group()
+                .add_argument(2, 3, 4, a=5, b=6, c=7),
+            mock.call.add_argument(9, 0, 1, a=2, b=3, c=4),
+            mock.call.add_mutually_exclusive_group(required=True),
+            mock.call.add_mutually_exclusive_group()
+                .add_argument(3, 4, 5, a=6, b=7, c=8),
+            mock.call.add_mutually_exclusive_group()
+                .add_argument(4, 5, 6, a=7, b=8, c=9),
+            mock.call.add_argument(0, 1, 2, a=3, b=4, c=5),
+            mock.call.add_subparsers(a=1, b=2, c=3),
+            mock.call.add_subparsers().add_parser(
+                prog='cmd_prog',
+                usage='cmd_usage',
+                description='cmd_description',
+                epilog='cmd_epilog',
+                formatter_class='cmd_formatter_class',
+            ),
+            mock.call.add_subparsers().add_parser(
+                prog='dmc_prog',
+                usage='dmc_usage',
+                description='dmc_description',
+                epilog='dmc_epilog',
+                formatter_class='dmc_formatter_class',
+            ),
+        ])
+        cmd_adaptor.setup_args.assert_called_once_with(cmd_parser)
+        dmc_adaptor.setup_args.assert_called_once_with(dmc_parser)
+        cmd_parser.set_defaults.assert_called_once_with(**{
+            sa._subcmd_attr: cmd_adaptor,
+        })
+        dmc_parser.set_defaults.assert_called_once_with(**{
+            sa._subcmd_attr: dmc_adaptor,
+        })
+
     @mock.patch.object(inspect, 'getargspec', return_value=inspect.ArgSpec(
         ('a', 'b', 'c'), None, None, None))
     def test_get_kwargs(self, mock_getargspec):
@@ -555,256 +770,305 @@ class TestScriptAdaptor(unittest2.TestCase):
         self.assertEqual(result, dict(a=1, b=2))
         mock_getargspec.assert_called_once_with(func)
 
-    def test_safe_call(self):
+    @mock.patch('sys.exc_info')
+    @mock.patch.object(inspect, 'isgeneratorfunction', return_value=False)
+    @mock.patch.object(cli_tools.ScriptAdaptor, 'get_kwargs',
+                       return_value=dict(a=1, b=2, c=3))
+    def test_safe_call(self, mock_get_kwargs, mock_isgeneratorfunction,
+                       mock_exc_info):
         func = mock.Mock(__doc__='', return_value='result')
         sa = cli_tools.ScriptAdaptor(func)
+        args = 'args'
 
-        result = sa.safe_call(dict(a=1, b=2, c=3))
+        result = sa.safe_call(args)
 
         self.assertEqual(result, ('result', None))
+        mock_get_kwargs.assert_called_once_with('args')
         func.assert_called_once_with(a=1, b=2, c=3)
 
-    @mock.patch('sys.exc_info', return_value='exception info')
-    def test_safe_call_noargs(self, mock_exc_info):
-        func = mock.Mock(__doc__='', side_effect=TestException("testing"))
+    @mock.patch('sys.exc_info')
+    @mock.patch.object(inspect, 'isgeneratorfunction', return_value=False)
+    @mock.patch.object(cli_tools.ScriptAdaptor, 'get_kwargs',
+                       return_value=dict(a=1, b=2, c=3))
+    def test_safe_call(self, mock_get_kwargs, mock_isgeneratorfunction,
+                       mock_exc_info):
+        func = mock.Mock(__doc__='', return_value='result')
         sa = cli_tools.ScriptAdaptor(func)
+        args = mock.Mock(debug=False)
 
-        result = sa.safe_call(dict(a=1, b=2, c=3))
+        result = sa.safe_call(args)
 
-        self.assertEqual(result, (None, 'exception info'))
+        self.assertEqual(result, ('result', None))
+        mock_get_kwargs.assert_called_once_with(args)
         func.assert_called_once_with(a=1, b=2, c=3)
 
-    def test_safe_call_exception_withdebug(self):
-        func = mock.Mock(__doc__='', side_effect=TestException("testing"))
+    @mock.patch('sys.exc_info', return_value=('type', 'exception', 'tb'))
+    @mock.patch.object(inspect, 'isgeneratorfunction', return_value=False)
+    @mock.patch.object(cli_tools.ScriptAdaptor, 'get_kwargs',
+                       return_value=dict(a=1, b=2, c=3))
+    def test_safe_call_exc(self, mock_get_kwargs, mock_isgeneratorfunction,
+                           mock_exc_info):
+        func = mock.Mock(__doc__='', side_effect=TestException)
         sa = cli_tools.ScriptAdaptor(func)
+        args = mock.Mock(debug=False)
 
-        self.assertRaises(TestException, sa.safe_call, dict(a=1, b=2, c=3),
-                          argparse.Namespace(debug=True))
+        result = sa.safe_call(args)
+
+        self.assertEqual(result, (None, ('type', 'exception', 'tb')))
+        mock_get_kwargs.assert_called_once_with(args)
         func.assert_called_once_with(a=1, b=2, c=3)
 
     @mock.patch.object(inspect, 'isgeneratorfunction', return_value=False)
-    @mock.patch.object(argparse, 'ArgumentParser', return_value=mock.Mock(**{
-        'parse_args.return_value': 'parsed args',
-    }))
-    @mock.patch.object(cli_tools.ScriptAdaptor, 'setup_args')
     @mock.patch.object(cli_tools.ScriptAdaptor, 'get_kwargs',
-                       return_value='keyword args')
-    @mock.patch.object(cli_tools.ScriptAdaptor, 'safe_call',
-                       return_value=('result', None))
-    def test_console_basic(self, mock_safe_call, mock_get_kwargs,
-                           mock_setup_args, mock_ArgumentParser,
-                           mock_isgeneratorfunction):
-        func = mock.Mock(__doc__='')
+                       return_value=dict(a=1, b=2, c=3))
+    def test_safe_call_exc_debug(self, mock_get_kwargs,
+                                 mock_isgeneratorfunction):
+        func = mock.Mock(__doc__='', side_effect=TestException)
         sa = cli_tools.ScriptAdaptor(func)
-        sa.prog = 'program'
-        sa.usage = 'usage'
-        sa.description = 'description'
-        sa.epilog = 'epilog'
-        sa.formatter_class = 'formatter_class'
+        args = mock.Mock(debug=True)
 
-        result = sa.console(argv='argument vector')
-
-        mock_ArgumentParser.assert_called_once_with(
-            prog='program', usage='usage', description='description',
-            epilog='epilog', formatter_class='formatter_class')
-        mock_setup_args.assert_called_once_with(
-            mock_ArgumentParser.return_value)
-        mock_ArgumentParser.return_value.parse_args.assert_called_once_with(
-            args='argument vector')
-        mock_get_kwargs.assert_called_once_with('parsed args')
-        mock_safe_call.assert_called_once_with('keyword args', 'parsed args')
-        self.assertEqual(result, 'result')
+        self.assertRaises(TestException, sa.safe_call, args)
+        mock_get_kwargs.assert_called_once_with(args)
+        func.assert_called_once_with(a=1, b=2, c=3)
 
+    @mock.patch('sys.exc_info')
     @mock.patch.object(inspect, 'isgeneratorfunction', return_value=False)
-    @mock.patch.object(argparse, 'ArgumentParser', return_value=mock.Mock(**{
-        'parse_args.return_value': 'parsed args',
-    }))
-    @mock.patch.object(cli_tools.ScriptAdaptor, 'setup_args')
     @mock.patch.object(cli_tools.ScriptAdaptor, 'get_kwargs',
-                       return_value='keyword args')
-    @mock.patch.object(cli_tools.ScriptAdaptor, 'safe_call',
-                       return_value=(None, ('type', 'exception', 'tb')))
-    def test_console_exception(self, mock_safe_call, mock_get_kwargs,
-                               mock_setup_args, mock_ArgumentParser,
-                               mock_isgeneratorfunction):
-        func = mock.Mock(__doc__='')
-        sa = cli_tools.ScriptAdaptor(func)
-        sa.prog = 'program'
-        sa.usage = 'usage'
-        sa.description = 'description'
-        sa.epilog = 'epilog'
-        sa.formatter_class = 'formatter_class'
-
-        result = sa.console(argv='argument vector')
-
-        mock_ArgumentParser.assert_called_once_with(
-            prog='program', usage='usage', description='description',
-            epilog='epilog', formatter_class='formatter_class')
-        mock_setup_args.assert_called_once_with(
-            mock_ArgumentParser.return_value)
-        mock_ArgumentParser.return_value.parse_args.assert_called_once_with(
-            args='argument vector')
-        mock_get_kwargs.assert_called_once_with('parsed args')
-        mock_safe_call.assert_called_once_with('keyword args', 'parsed args')
-        self.assertEqual(result, 'exception')
-
-    @mock.patch.object(inspect, 'isgeneratorfunction', return_value=False)
-    @mock.patch.object(argparse, 'ArgumentParser', return_value=mock.Mock(**{
-        'parse_args.return_value': 'parsed args',
-    }))
-    @mock.patch.object(cli_tools.ScriptAdaptor, 'setup_args')
-    @mock.patch.object(cli_tools.ScriptAdaptor, 'get_kwargs',
-                       return_value='keyword args')
-    @mock.patch.object(cli_tools.ScriptAdaptor, 'safe_call',
-                       return_value=('result', None))
-    def test_console_processor_func(self, mock_safe_call, mock_get_kwargs,
-                                    mock_setup_args, mock_ArgumentParser,
-                                    mock_isgeneratorfunction):
-        func = mock.Mock(__doc__='')
+                       return_value=dict(a=1, b=2, c=3))
+    def test_safe_call_proc_func(self, mock_get_kwargs,
+                                 mock_isgeneratorfunction, mock_exc_info):
+        func = mock.Mock(__doc__='', return_value='result')
         sa = cli_tools.ScriptAdaptor(func)
         sa._processor = mock.Mock()
+        args = mock.Mock(debug=False)
 
-        result = sa.console('override args')
+        result = sa.safe_call(args)
 
-        self.assertFalse(mock_ArgumentParser.called)
-        sa._processor.assert_called_once_with('override args')
-        mock_get_kwargs.assert_called_once_with('override args')
-        mock_safe_call.assert_called_once_with('keyword args', 'override args')
-        self.assertEqual(result, 'result')
+        self.assertEqual(result, ('result', None))
+        sa._processor.assert_called_once_with(args)
+        mock_get_kwargs.assert_called_once_with(args)
+        func.assert_called_once_with(a=1, b=2, c=3)
 
+    @mock.patch('sys.exc_info')
     @mock.patch.object(inspect, 'isgeneratorfunction', return_value=True)
-    @mock.patch.object(argparse, 'ArgumentParser', return_value=mock.Mock(**{
-        'parse_args.return_value': 'parsed args',
-    }))
-    @mock.patch.object(cli_tools.ScriptAdaptor, 'setup_args')
     @mock.patch.object(cli_tools.ScriptAdaptor, 'get_kwargs',
-                       return_value='keyword args')
-    @mock.patch.object(cli_tools.ScriptAdaptor, 'safe_call',
-                       return_value=('result', None))
-    def test_console_processor_gen_nopost(self, mock_safe_call,
-                                          mock_get_kwargs, mock_setup_args,
-                                          mock_ArgumentParser,
-                                          mock_isgeneratorfunction):
-        func = mock.Mock(__doc__='')
+                       return_value=dict(a=1, b=2, c=3))
+    def test_safe_call_proc_gen_nopost(self, mock_get_kwargs,
+                                       mock_isgeneratorfunction,
+                                       mock_exc_info):
+        func = mock.Mock(__doc__='', return_value='result')
         sa = cli_tools.ScriptAdaptor(func)
-        generator = mock.Mock(**{
+        sa._processor = mock.Mock(return_value=mock.Mock(**{
             'next.side_effect': StopIteration,
-        })
-        sa._processor = mock.Mock(return_value=generator)
+        }))
+        args = mock.Mock(debug=False)
 
-        result = sa.console('override args')
+        result = sa.safe_call(args)
 
-        self.assertFalse(mock_ArgumentParser.called)
-        sa._processor.assert_called_once_with('override args')
-        generator.assert_has_calls([
+        self.assertEqual(result, ('result', None))
+        sa._processor.assert_called_once_with(args)
+        sa._processor.return_value.assert_has_calls([
             mock.call.next(),
         ])
-        self.assertEqual(len(generator.method_calls), 1)
-        mock_get_kwargs.assert_called_once_with('override args')
-        mock_safe_call.assert_called_once_with('keyword args', 'override args')
-        self.assertEqual(result, 'result')
+        self.assertEqual(len(sa._processor.return_value.method_calls), 1)
+        mock_get_kwargs.assert_called_once_with(args)
+        func.assert_called_once_with(a=1, b=2, c=3)
 
+    @mock.patch('sys.exc_info')
     @mock.patch.object(inspect, 'isgeneratorfunction', return_value=True)
-    @mock.patch.object(argparse, 'ArgumentParser', return_value=mock.Mock(**{
-        'parse_args.return_value': 'parsed args',
-    }))
-    @mock.patch.object(cli_tools.ScriptAdaptor, 'setup_args')
     @mock.patch.object(cli_tools.ScriptAdaptor, 'get_kwargs',
-                       return_value='keyword args')
-    @mock.patch.object(cli_tools.ScriptAdaptor, 'safe_call',
-                       return_value=('result', None))
-    def test_console_processor_gen_override(self, mock_safe_call,
-                                            mock_get_kwargs, mock_setup_args,
-                                            mock_ArgumentParser,
-                                            mock_isgeneratorfunction):
-        func = mock.Mock(__doc__='')
+                       return_value=dict(a=1, b=2, c=3))
+    def test_safe_call_proc_gen_post_res_noreplace(self, mock_get_kwargs,
+                                                   mock_isgeneratorfunction,
+                                                   mock_exc_info):
+        func = mock.Mock(__doc__='', return_value='result')
         sa = cli_tools.ScriptAdaptor(func)
-        generator = mock.Mock(**{
-            'send.return_value': 'override result',
-        })
-        sa._processor = mock.Mock(return_value=generator)
-
-        result = sa.console('override args')
-
-        self.assertFalse(mock_ArgumentParser.called)
-        sa._processor.assert_called_once_with('override args')
-        generator.assert_has_calls([
-            mock.call.next(),
-            mock.call.send('result'),
-            mock.call.close(),
-        ])
-        self.assertEqual(len(generator.method_calls), 3)
-        mock_get_kwargs.assert_called_once_with('override args')
-        mock_safe_call.assert_called_once_with('keyword args', 'override args')
-        self.assertEqual(result, 'override result')
-
-    @mock.patch.object(inspect, 'isgeneratorfunction', return_value=True)
-    @mock.patch.object(argparse, 'ArgumentParser', return_value=mock.Mock(**{
-        'parse_args.return_value': 'parsed args',
-    }))
-    @mock.patch.object(cli_tools.ScriptAdaptor, 'setup_args')
-    @mock.patch.object(cli_tools.ScriptAdaptor, 'get_kwargs',
-                       return_value='keyword args')
-    @mock.patch.object(cli_tools.ScriptAdaptor, 'safe_call',
-                       return_value=('result', None))
-    def test_console_processor_gen_no_override(self, mock_safe_call,
-                                               mock_get_kwargs,
-                                               mock_setup_args,
-                                               mock_ArgumentParser,
-                                               mock_isgeneratorfunction):
-        func = mock.Mock(__doc__='')
-        sa = cli_tools.ScriptAdaptor(func)
-        generator = mock.Mock(**{
+        sa._processor = mock.Mock(return_value=mock.Mock(**{
             'send.side_effect': StopIteration,
-        })
-        sa._processor = mock.Mock(return_value=generator)
+        }))
+        args = mock.Mock(debug=False)
 
-        result = sa.console('override args')
+        result = sa.safe_call(args)
 
-        self.assertFalse(mock_ArgumentParser.called)
-        sa._processor.assert_called_once_with('override args')
-        generator.assert_has_calls([
+        self.assertEqual(result, ('result', None))
+        sa._processor.assert_called_once_with(args)
+        sa._processor.return_value.assert_has_calls([
             mock.call.next(),
             mock.call.send('result'),
             mock.call.close(),
         ])
-        self.assertEqual(len(generator.method_calls), 3)
-        mock_get_kwargs.assert_called_once_with('override args')
-        mock_safe_call.assert_called_once_with('keyword args', 'override args')
-        self.assertEqual(result, 'result')
+        self.assertEqual(len(sa._processor.return_value.method_calls), 3)
+        mock_get_kwargs.assert_called_once_with(args)
+        func.assert_called_once_with(a=1, b=2, c=3)
 
+    @mock.patch('sys.exc_info')
     @mock.patch.object(inspect, 'isgeneratorfunction', return_value=True)
-    @mock.patch.object(argparse, 'ArgumentParser', return_value=mock.Mock(**{
-        'parse_args.return_value': 'parsed args',
-    }))
-    @mock.patch.object(cli_tools.ScriptAdaptor, 'setup_args')
     @mock.patch.object(cli_tools.ScriptAdaptor, 'get_kwargs',
-                       return_value='keyword args')
-    @mock.patch.object(cli_tools.ScriptAdaptor, 'safe_call',
-                       return_value=(None, ('type', 'exception', 'tb')))
-    def test_console_processor_gen_throw(self, mock_safe_call, mock_get_kwargs,
-                                         mock_setup_args, mock_ArgumentParser,
-                                         mock_isgeneratorfunction):
-        func = mock.Mock(__doc__='')
+                       return_value=dict(a=1, b=2, c=3))
+    def test_safe_call_proc_gen_post_res_replace(self, mock_get_kwargs,
+                                                 mock_isgeneratorfunction,
+                                                 mock_exc_info):
+        func = mock.Mock(__doc__='', return_value='result')
         sa = cli_tools.ScriptAdaptor(func)
-        generator = mock.Mock(**{
-            'throw.side_effect': TestException("testing"),
-        })
-        sa._processor = mock.Mock(return_value=generator)
+        sa._processor = mock.Mock(return_value=mock.Mock(**{
+            'send.return_value': 'override',
+        }))
+        args = mock.Mock(debug=False)
 
-        result = sa.console('override args')
+        result = sa.safe_call(args)
 
-        self.assertFalse(mock_ArgumentParser.called)
-        sa._processor.assert_called_once_with('override args')
-        generator.assert_has_calls([
+        self.assertEqual(result, ('override', None))
+        sa._processor.assert_called_once_with(args)
+        sa._processor.return_value.assert_has_calls([
+            mock.call.next(),
+            mock.call.send('result'),
+            mock.call.close(),
+        ])
+        self.assertEqual(len(sa._processor.return_value.method_calls), 3)
+        mock_get_kwargs.assert_called_once_with(args)
+        func.assert_called_once_with(a=1, b=2, c=3)
+
+    @mock.patch('sys.exc_info', return_value=('type', 'exception', 'tb'))
+    @mock.patch.object(inspect, 'isgeneratorfunction', return_value=True)
+    @mock.patch.object(cli_tools.ScriptAdaptor, 'get_kwargs',
+                       return_value=dict(a=1, b=2, c=3))
+    def test_safe_call_proc_gen_post_exc_noreplace(self, mock_get_kwargs,
+                                                   mock_isgeneratorfunction,
+                                                   mock_exc_info):
+        func = mock.Mock(__doc__='', side_effect=TestException)
+        sa = cli_tools.ScriptAdaptor(func)
+        sa._processor = mock.Mock(return_value=mock.Mock(**{
+            'throw.return_value': 'thrown',
+        }))
+        args = mock.Mock(debug=False)
+
+        result = sa.safe_call(args)
+
+        self.assertEqual(result, ('thrown', None))
+        sa._processor.assert_called_once_with(args)
+        sa._processor.return_value.assert_has_calls([
             mock.call.next(),
             mock.call.throw('type', 'exception', 'tb'),
             mock.call.close(),
         ])
-        self.assertEqual(len(generator.method_calls), 3)
-        mock_get_kwargs.assert_called_once_with('override args')
-        mock_safe_call.assert_called_once_with('keyword args', 'override args')
-        self.assertEqual(result, 'testing')
+        self.assertEqual(len(sa._processor.return_value.method_calls), 3)
+        mock_get_kwargs.assert_called_once_with(args)
+        func.assert_called_once_with(a=1, b=2, c=3)
+
+    @mock.patch('sys.exc_info', side_effect=[
+        ('type', 'exception', 'tb'),
+        ('otype', 'something', 'bt'),
+    ])
+    @mock.patch.object(inspect, 'isgeneratorfunction', return_value=True)
+    @mock.patch.object(cli_tools.ScriptAdaptor, 'get_kwargs',
+                       return_value=dict(a=1, b=2, c=3))
+    def test_safe_call_proc_gen_post_exc_replace(self, mock_get_kwargs,
+                                                 mock_isgeneratorfunction,
+                                                 mock_exc_info):
+        func = mock.Mock(__doc__='', side_effect=TestException)
+        sa = cli_tools.ScriptAdaptor(func)
+        sa._processor = mock.Mock(return_value=mock.Mock(**{
+            'throw.side_effect': TestException,
+        }))
+        args = mock.Mock(debug=False)
+
+        result = sa.safe_call(args)
+
+        self.assertEqual(result, (None, ('otype', 'something', 'bt')))
+        sa._processor.assert_called_once_with(args)
+        sa._processor.return_value.assert_has_calls([
+            mock.call.next(),
+            mock.call.throw('type', 'exception', 'tb'),
+            mock.call.close(),
+        ])
+        self.assertEqual(len(sa._processor.return_value.method_calls), 3)
+        mock_get_kwargs.assert_called_once_with(args)
+        func.assert_called_once_with(a=1, b=2, c=3)
+
+    @mock.patch.object(argparse, 'ArgumentParser', return_value=mock.Mock(**{
+        'parse_args.return_value': 'parsed args',
+    }))
+    @mock.patch.object(cli_tools.ScriptAdaptor, 'setup_args')
+    @mock.patch.object(cli_tools.ScriptAdaptor, 'safe_call',
+                       return_value=('result', None))
+    def test_console_basic(self, mock_safe_call, mock_setup_args,
+                           mock_ArgumentParser):
+        func = mock.Mock(__doc__='')
+        sa = cli_tools.ScriptAdaptor(func)
+        sa.prog = 'program'
+        sa.usage = 'usage'
+        sa.description = 'description'
+        sa.epilog = 'epilog'
+        sa.formatter_class = 'formatter_class'
+
+        result = sa.console(argv='argument vector')
+
+        mock_ArgumentParser.assert_called_once_with(
+            prog='program', usage='usage', description='description',
+            epilog='epilog', formatter_class='formatter_class')
+        mock_setup_args.assert_called_once_with(
+            mock_ArgumentParser.return_value)
+        mock_ArgumentParser.return_value.parse_args.assert_called_once_with(
+            args='argument vector')
+        mock_safe_call.assert_called_once_with('parsed args')
+        self.assertEqual(result, 'result')
+
+    @mock.patch.object(argparse, 'ArgumentParser', return_value=mock.Mock(**{
+        'parse_args.return_value': 'parsed args',
+    }))
+    @mock.patch.object(cli_tools.ScriptAdaptor, 'setup_args')
+    @mock.patch.object(cli_tools.ScriptAdaptor, 'safe_call',
+                       return_value=(None, ('type', 'exception', 'tb')))
+    def test_console_exception(self, mock_safe_call, mock_setup_args,
+                               mock_ArgumentParser):
+        func = mock.Mock(__doc__='')
+        sa = cli_tools.ScriptAdaptor(func)
+        sa.prog = 'program'
+        sa.usage = 'usage'
+        sa.description = 'description'
+        sa.epilog = 'epilog'
+        sa.formatter_class = 'formatter_class'
+
+        result = sa.console(args='override args')
+
+        self.assertFalse(mock_ArgumentParser.called)
+        self.assertFalse(mock_setup_args.called)
+        mock_safe_call.assert_called_once_with('override args')
+        self.assertEqual(result, 'exception')
+
+    @mock.patch.object(argparse, 'ArgumentParser', return_value=mock.Mock(**{
+        'parse_args.return_value': 'parsed args',
+    }))
+    @mock.patch.object(cli_tools.ScriptAdaptor, 'setup_args')
+    @mock.patch.object(cli_tools.ScriptAdaptor, 'safe_call',
+                       return_value=('result', None))
+    def test_console_subcmd(self, mock_safe_call, mock_setup_args,
+                            mock_ArgumentParser):
+        func = mock.Mock(__doc__='')
+        sa = cli_tools.ScriptAdaptor(func)
+        sa.do_subs = True
+        adaptor = mock.Mock(**{'safe_call.return_value': ('tluser', None)})
+        args = mock.Mock(**{sa._subcmd_attr: adaptor})
+
+        result = sa.console(args=args)
+
+        self.assertFalse(mock_ArgumentParser.called)
+        self.assertFalse(mock_setup_args.called)
+        self.assertFalse(mock_safe_call.called)
+        adaptor.safe_call.assert_called_once_with(args)
+        self.assertEqual(result, 'tluser')
+
+    def test_get_subcommands(self):
+        func = mock.Mock(__doc__='')
+        sa = cli_tools.ScriptAdaptor(func)
+        sa._subcommands = mock.Mock(**{'items.return_value': [
+            ('cmd', mock.Mock(_func='subcmd')),
+            ('dmc', mock.Mock(_func='subdmc')),
+        ]})
+
+        result = sa.get_subcommands()
+
+        self.assertEqual(result, dict(cmd='subcmd', dmc='subdmc'))
 
 
 class TestDecorators(unittest2.TestCase):
@@ -832,10 +1096,9 @@ class TestDecorators(unittest2.TestCase):
         self.assertEqual(result, func)
         self.assertEqual(mock_get_adaptor.return_value.prog, 'program')
 
-    @mock.patch.object(cli_tools, '_clean_text', side_effect=lambda x: x)
     @mock.patch.object(cli_tools.ScriptAdaptor, '_get_adaptor',
                        return_value=mock.Mock())
-    def test_usage(self, mock_get_adaptor, mock_clean_text):
+    def test_usage(self, mock_get_adaptor):
         decorator = cli_tools.usage('text')
 
         self.assertTrue(callable(decorator))
@@ -845,14 +1108,12 @@ class TestDecorators(unittest2.TestCase):
         result = decorator(func)
 
         mock_get_adaptor.assert_called_once_with(func)
-        mock_clean_text.assert_called_once_with('text')
         self.assertEqual(result, func)
         self.assertEqual(mock_get_adaptor.return_value.usage, 'text')
 
-    @mock.patch.object(cli_tools, '_clean_text', side_effect=lambda x: x)
     @mock.patch.object(cli_tools.ScriptAdaptor, '_get_adaptor',
                        return_value=mock.Mock())
-    def test_description(self, mock_get_adaptor, mock_clean_text):
+    def test_description(self, mock_get_adaptor):
         decorator = cli_tools.description('text')
 
         self.assertTrue(callable(decorator))
@@ -862,14 +1123,12 @@ class TestDecorators(unittest2.TestCase):
         result = decorator(func)
 
         mock_get_adaptor.assert_called_once_with(func)
-        mock_clean_text.assert_called_once_with('text')
         self.assertEqual(result, func)
         self.assertEqual(mock_get_adaptor.return_value.description, 'text')
 
-    @mock.patch.object(cli_tools, '_clean_text', side_effect=lambda x: x)
     @mock.patch.object(cli_tools.ScriptAdaptor, '_get_adaptor',
                        return_value=mock.Mock())
-    def test_epilog(self, mock_get_adaptor, mock_clean_text):
+    def test_epilog(self, mock_get_adaptor):
         decorator = cli_tools.epilog('text')
 
         self.assertTrue(callable(decorator))
@@ -879,7 +1138,6 @@ class TestDecorators(unittest2.TestCase):
         result = decorator(func)
 
         mock_get_adaptor.assert_called_once_with(func)
-        mock_clean_text.assert_called_once_with('text')
         self.assertEqual(result, func)
         self.assertEqual(mock_get_adaptor.return_value.epilog, 'text')
 
@@ -962,3 +1220,36 @@ class TestDecorators(unittest2.TestCase):
         self.assertEqual(result, func)
         mock_get_adaptor.return_value._add_group.assert_called_once_with(
             'group', 'exclusive', dict(a=1, b=2, c=3))
+
+    @mock.patch.object(cli_tools.ScriptAdaptor, '_get_adaptor',
+                       return_value=mock.Mock())
+    def test_subparsers(self, mock_get_adaptor):
+        decorator = cli_tools.subparsers(a=1, b=2, c=3)
+
+        self.assertTrue(callable(decorator))
+        self.assertFalse(mock_get_adaptor.called)
+
+        func = mock.Mock()
+        result = decorator(func)
+
+        mock_get_adaptor.assert_called_once_with(func)
+        self.assertEqual(result, func)
+        self.assertEqual(mock_get_adaptor.return_value.subkwargs,
+                         dict(a=1, b=2, c=3))
+        self.assertEqual(mock_get_adaptor.return_value.do_subs, True)
+
+    @mock.patch.object(cli_tools.ScriptAdaptor, '_get_adaptor',
+                       return_value=mock.Mock())
+    def test_load_subcommands(self, mock_get_adaptor):
+        decorator = cli_tools.load_subcommands('entrypoint.group')
+
+        self.assertTrue(callable(decorator))
+        self.assertFalse(mock_get_adaptor.called)
+
+        func = mock.Mock()
+        result = decorator(func)
+
+        mock_get_adaptor.assert_called_once_with(func)
+        self.assertEqual(result, func)
+        mock_get_adaptor.return_value._add_extensions.assert_called_once_with(
+            'entrypoint.group')
